import functools
import logging
import time
from typing import Any, Callable, Tuple, Type


def retry_if_exception(
    exceptions: Tuple[Type[BaseException], ...] = (),
    max_retries: int = 10,
    delay: float = 2.0,
    failure_return: Any = None,
) -> Callable:
    """
    Decorator thử lại một hàm nếu nó gây ra các ngoại lệ được chỉ định.
    Args:
        exceptions: Tuple các loại ngoại lệ để bắt và thử lại
        max_retries: Số lần thử lại tối đa
        delay: Thời gian chờ ban đầu giữa các lần thử lại (giây)
        failure_return: Giá trị trả về nếu thất bại
        message_on_retry: Message khi retry (default: RETRY: <function_name>)
    """

    def wrapper(func):
        @functools.wraps(func)
        def method(self, *args, **kwargs) -> Any:
            logger: logging.Logger = (
                getattr(self, "logger")
                if hasattr(self, "logger") and isinstance(getattr(self,'logger'),logging.Logger)
                else logging.getLogger()
            )
            retries = 0
            while True:
                try:
                    return func(self, *args, **kwargs)
                except exceptions as e:
                    retries += 1
                    if retries > max_retries:
                        msg = e.msg if hasattr(e, "msg") else str(e)
                        logger.error(
                            f"{func.__name__} failed after {max_retries} retries: {msg}"
                        )
                        return failure_return
                    logger.info(f"RETRY {func.__name__}")
                    time.sleep(delay)
                except Exception as e:
                    msg = e.msg if hasattr(e, "msg") else str(e)
                    logger.error(f"{func.__name__} catch {type(e)}: {msg} ")
                    return failure_return

        return method

    return wrapper


import os
import re
import time
from datetime import date
from typing import List
from urllib.parse import urljoin

import pandas as pd
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select

from NSK.bot.base import IBot, IBotMeta
from NSK.common import retry_if_exception


class WebAccessMeta(IBotMeta):
    def __new__(cls, name, bases, class_dict):
        for attr_name, attr_value in class_dict.items():
            if callable(attr_value) and attr_name not in [
                "__init__",
                "_authentication",
            ]:
                class_dict[attr_name] = cls.login(attr_value)
        return super().__new__(cls, name, bases, class_dict)

    @staticmethod
    def login(func):
        def wrapper(self, *args, **kwargs):
            if not self.authenticated:
                raise ConnectionRefusedError("Yêu cầu xác thực")
            if (
                self.browser.current_url.endswith(".com/")
                or self.browser.current_url == "data:,"
            ):
                self.authenticated = self._authentication(
                    self._username, self._password
                )
            return func(self, *args, **kwargs)

        return wrapper


class WebAccess(IBot, metaclass=WebAccessMeta):
    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.url = url
        self._username = username
        self._password = password
        self.authenticated = self._authentication(self._username, self._password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")

    @retry_if_exception(
        delay=2,
        failure_return=False,
    )
    def _authentication(self, username: str, password: str) -> bool:
        self.navigate(self.url)
        self.wait.until(
            method = EC.element_to_be_clickable((By.CSS_SELECTOR, "input[type='text']")),
            message= "Không tìm thấy usernameField",
        ).send_keys(username)
        self.wait.until(
            method  = EC.element_to_be_clickable((By.CSS_SELECTOR, "input[type='password']")),
            message = "Không tìm thấy passwordField",
        ).send_keys(password)
        self.wait.until(
            method  = EC.element_to_be_clickable((By.CSS_SELECTOR, "button[class='btn login']")),
            message = "Không tìm thấy loginButton",
        ).click()
        try:
            error_box = self.wait.until(
                method=EC.presence_of_element_located((By.CSS_SELECTOR, "div[id='f-error-box']"))
            )
            data = error_box.find_element(By.CSS_SELECTOR, "div[class='data']")
            self.logger.info(f"Xác thực thất bại!: {data.text}")
            return False
        except TimeoutException:
            self.logger.info("Authenticated")
            return True

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
        ),
    )
    def update(self, case_number: int | str, **kwargs) -> bool:
        self.logger.info(f"Update {case_number}: {kwargs}")
        # --- Redirect --- #
        f_menus = self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "div[id='f-menus']"))
        )
        if order_list := f_menus.find_elements(By.CSS_SELECTOR, "a[title='受注一覧']"):
            href = order_list[0].get_attribute("href")
            self.navigate(href)
        else:
            raise LookupError("Not Found 'Order Tab'")
        # Clear
        clear_btn = self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "button[type='reset']"))
        )
        self.browser.execute_script(
            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
            clear_btn,
        )
        time.sleep(self.retry_interval)
        self.wait.until(EC.element_to_be_clickable(clear_btn)).click()
        # Input Case Number
        case_number_td = self.wait.until(
            EC.presence_of_element_located(
                (By.XPATH, "//th[span[text()='案件番号']]/following-sibling::td")
            )
        )
        self.browser.execute_script(
            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
            case_number_td,
        )
        time.sleep(self.retry_interval)
        input_field = case_number_td.find_element(By.TAG_NAME, "input")
        self.wait.until(EC.element_to_be_clickable(input_field)).send_keys(case_number)
        # Search
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "button[type='submit']"))
        ).click()
        time.sleep(self.retry_interval)
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        table = self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "table[id='orderlist']"))
        )
        # Data in tbody
        tbody = table.find_element(By.TAG_NAME, "tbody")
        trs = tbody.find_elements(By.TAG_NAME, "tr")
        if len(trs) == 0:
            raise LookupError(f"Not Found Case Number: {case_number}")
        if trs[0].text == "検索結果はありません":
            raise LookupError(f"Not Found Case Number: {case_number}")
        hrefs = []
        for tr in trs:
            refer_btn = tr.find_element(By.CSS_SELECTOR, 'input[value="参照"]')
            self.browser.execute_script(
                "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                refer_btn,
            )
            time.sleep(self.retry_interval)
            self.wait.until(EC.element_to_be_clickable(refer_btn))
            onclick = refer_btn.get_attribute("onclick")
            match = re.search(r"location\.href='([^']+)'", onclick)
            if not match:
                raise RuntimeError(f"Update {case_number} failed")
            hrefs.append(urljoin(self.url, match.group(1)))
        time.sleep(self.retry_interval * 2)
        for href in hrefs:
            self.navigate(href)
            # --- Serializer Kwargs --- #
            for key in kwargs.keys():
                if not isinstance(kwargs.get(key), list):
                    kwargs[key] = [kwargs.get(key)]
            # --- Update Web Access --- #
            for key in kwargs.keys():
                try:
                    ths = self.wait.until(
                        EC.presence_of_all_elements_located(
                            (By.XPATH, f"//th[text()='{key}']")
                        )
                    )[:len(kwargs.get(key,None))]
                    for insert_index, th in enumerate(ths):
                        self.browser.execute_script(
                            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                            th,
                        )
                        time.sleep(self.retry_interval)
                        table = th.find_element(By.XPATH, "./ancestor::table[1]")
                        tbody = table.find_element(By.TAG_NAME, "tbody")
                        start_index = None  # Cột Header
                        end_index = None  # Cột Header Tiếp theo
                        for index, tr in enumerate(
                            tbody.find_elements(By.TAG_NAME, "tr")
                        ):
                            if tr.find_elements(By.XPATH, f".//th[text()='{key}']"):
                                start_index = index
                                break
                        if start_index is None:
                            raise LookupError("Không tìm được vị trí dòng dữ liệu!")
                        for index, tr in enumerate(
                            tbody.find_elements(By.TAG_NAME, "tr")[start_index + 1 :]
                        ):
                            if tr.find_elements(By.CSS_SELECTOR, "th[class='c']"):
                                end_index = index + start_index + 1
                                break
                        table = tbody.find_elements(By.TAG_NAME, "tr")[
                            start_index:end_index
                        ]
                        header = table[0]
                        pos_index = None
                        for index, column in enumerate(
                            header.find_elements(By.TAG_NAME, "th")
                        ):
                            if column.text == key:
                                pos_index = index
                                break
                        if pos_index is None:
                            raise LookupError("No Data Column")
                        for tr in tbody.find_elements(By.TAG_NAME, "tr")[
                            start_index + 1 : end_index
                        ]:
                            td = tr.find_elements(By.TAG_NAME, "td")[pos_index]
                            input_field = td.find_element(
                                By.XPATH, ".//input | .//select"
                            )
                            self.wait.until(
                                EC.element_to_be_clickable(input_field)
                            ).click()
                            self.browser.execute_script(
                                "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                                input_field,
                            )
                            time.sleep(self.retry_interval)
                            if input_field.tag_name == "input":
                                if (
                                    kwargs[key][insert_index].get("mode")
                                    == "replace"
                                ):
                                    input_field.clear()
                                    insert_data = kwargs[key][insert_index].get(
                                        "data"
                                    )
                                    input_field.send_keys(insert_data)
                                    input_field.send_keys(Keys.ENTER)
                                if (
                                    kwargs[key][insert_index].get("mode")
                                    == "append"
                                ):
                                    input_field.click()
                                    insert_data = (
                                        f" {kwargs[key][insert_index].get('data')}"
                                    )
                                    if not input_field.get_attribute("value").endswith(
                                        insert_data
                                    ):
                                        input_field.send_keys(insert_data)
                                        input_field.send_keys(Keys.ENTER)
                                if (
                                    kwargs[key][insert_index].get("mode")
                                    == "insert"
                                ):
                                    input_field.click()
                                    input_field.send_keys(Keys.HOME)
                                    insert_data = (
                                        f"{kwargs[key][insert_index].get('data')} "
                                    )
                                    if not input_field.get_attribute("value").startswith(
                                        insert_data
                                    ):
                                        input_field.send_keys(insert_data)
                                        input_field.send_keys(Keys.ENTER)
                            elif input_field.tag_name == "select":
                                select = Select(input_field)
                                select.select_by_visible_text(
                                    kwargs[key][insert_index].get("data")
                                )
                except TimeoutException:
                    pass
            time.sleep(self.retry_interval)
            update_btn = self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, "button[id='order_update']")
                )
            )
            self.browser.execute_script("window.scrollTo(0, 0);")
            time.sleep(self.retry_interval * 2)
            self.wait.until(EC.element_to_be_clickable(update_btn)).click()
            time.sleep(self.retry_interval)
            while (
                self.browser.execute_script("return document.readyState") != "complete"
            ):
                time.sleep(self.retry_interval)
            notice = self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, "div[id='f-notice-box']")
                )
            )
            self.logger.info(f"Update Case: {case_number}: {repr(notice.text)}")
            if notice.text != "案件情報を更新しました。\n閉じる":
                return False
        return True


    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            NotImplementedError,
        ),
        delay=2,
        failure_return=pd.DataFrame(),
    )
    def get_order_list(
        self,
        building_name: str | None = None,
        drawing: list[str] | None = None,
        delivery_date: list[str] = [date.today().strftime("%Y/%m/%d"), ""],
        fields: list[str] | None = None,
    ) -> pd.DataFrame:
        self.logger.info("Get Order List")
        f_menus = self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "div[id='f-menus']"))
        )
        if order_list := f_menus.find_elements(By.CSS_SELECTOR, "a[title='受注一覧']"):
            href = order_list[0].get_attribute("href")
            self.navigate(href)
        else:
            self.logger.error("Can't find Order List")
            return pd.DataFrame()
        # Clear
        self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "button[type='reset']"))
        )
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "button[type='reset']"))
        ).click()
        # Filter
        if building_name:
            self.logger.info(f"Filter Building: {building_name}")
            time.sleep(1)
            self.wait.until(
                EC.element_to_be_clickable(
                    (By.ID, "select2-search_builder_cd-container")
                )
            ).click()
            time.sleep(1)
            self.wait.until(
                EC.element_to_be_clickable(
                    (By.CSS_SELECTOR, 'input[class="select2-search__field"]')
                )
            ).send_keys(building_name)
            options: List[WebElement]= self.wait.until(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR, "ul#select2-search_builder_cd-results > li"))
            )
            if len(options) > 1:
                for option in options:
                    self.logger.info(f"Found: {option.text}")
                raise LookupError(f"Found {len(options)} options")
            option = options[0]
            if option.text == "No results found":
                raise LookupError("No results found")
            self.logger.info(f"Select: {option.text}")
            self.wait.until(EC.element_to_be_clickable(option)).click()
            time.sleep(1)
        if delivery_date:  # Delivery date
            self.logger.info(f"Filter Delivery Date from {delivery_date[0]} to {delivery_date[1]}")
            self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, 'input[name="search_fix_deliver_date_from"]')
                )
            ).clear()
            self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, 'input[name="search_fix_deliver_date_from"]')
                )
            ).send_keys(delivery_date[0])
            self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, 'input[name="search_fix_deliver_date_to"]'))
            ).clear()
            self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, 'input[name="search_fix_deliver_date_to"]')
                )
            ).send_keys(delivery_date[1])
            self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, 'input[name="search_fix_deliver_date_to"]')
                )
            ).send_keys(Keys.ESCAPE)
        if drawing:  # Drawing
            self.logger.info(f"Filter Drawing: {drawing}")
            self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "button[id='search_drawing_type_ms']"))
            )
            self.wait.until(
                EC.element_to_be_clickable(
                    (By.CSS_SELECTOR, "button[id='search_drawing_type_ms']")
                )
            ).click()
            for e in drawing:
                xpath = f"//span[text()='{e}']"
                self.wait.until(EC.presence_of_element_located((By.XPATH, xpath)))
                self.wait.until(EC.element_to_be_clickable((By.XPATH, xpath))).click()
            self.wait.until(
                EC.element_to_be_clickable(
                    (By.CSS_SELECTOR, "button[id='search_drawing_type_ms']")
                )
            ).send_keys(Keys.ESCAPE)
        # Search
        submit_btn = self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "button[type='submit']"))
        )
        self.browser.execute_script(
            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
            submit_btn,
        )
        time.sleep(self.retry_interval)
        self.wait.until(EC.element_to_be_clickable(submit_btn)).click()
        time.sleep(self.retry_interval)
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        # Download File
        download_btn = self.wait.until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, "a[class='button fa fa-download']")
            )
        )
        self.browser.execute_script(
            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
            download_btn,
        )
        time.sleep(self.retry_interval)
        self.wait.until(EC.element_to_be_clickable(download_btn)).click()
        time.sleep(self.timeout)
        file_name, tag = self.wait_for_download_to_finish()
        if tag:
            self.logger.error(tag)
            time.sleep(self.timeout)
            raise NotImplementedError(f"Lấy danh sách đơn hàng thất bại: {tag}")
        file_path = os.path.join(self.download_directory, file_name)
        df = pd.read_csv(file_path, encoding="CP932",low_memory=False)
        df = df[fields] if fields else df
        os.remove(file_path)
        self.logger.info("Get order list successfully")
        return df


__all__ = ['WebAccess']


import os
import re
import time
from typing import List, Tuple
from urllib.parse import unquote

import pandas as pd
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        NoSuchElementException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from NSK.bot.base import IBot
from NSK.common import retry_if_exception


class SharePoint(IBot):
    redundant_name = r'[\ue000-\uf8ff]|\n|Press C to open file hover card'

    def __init__(self, url: str, username: str, password: str, **kwargs):
        super().__init__(**kwargs)
        self.url = url
        self.authenticated = self._authentication(username, password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")

    def navigate(self, url, wait_for_complete = True):
        super().navigate(url, wait_for_complete)
        try:
            ms_error_header = self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"div[id='ms-error-header']"))
            )
            self.logger.error(ms_error_header.text)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"div[id='ms-error'] a"))
            ).click()
            time.sleep(self.retry_interval)
            while self.browser.execute_script("return document.readyState") != "complete":
                continue
            time.sleep(self.retry_interval)
            return self.navigate(url, wait_for_complete)
        except Exception:
            pass

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
            Exception,
        ),
        failure_return=False,
    )
    def _authentication(self, username: str, password: str) -> bool:
        self.navigate("https://login.microsoftonline.com/")
        time.sleep(self.retry_interval)
        if self.browser.current_url.startswith("https://m365.cloud.microsoft/?auth="):
            self.logger.info("Authenticated")
            return True
        # -- Username
        time.sleep(self.retry_interval)
        try:
            self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, 'input[type="email"]'))
            ).send_keys(username)
            # -- Next
            btn = self.wait.until(
                EC.presence_of_element_located((By.ID, "idSIButton9"))
            )
            self.wait.until(EC.element_to_be_clickable(btn)).click()
            # -- Password
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, 'input[type="password"]'))
            ).send_keys(password)
            # -- Sign in
            self.wait.until(EC.element_to_be_clickable((By.ID, "idSIButton9"))).click()
        except TimeoutException:
            alert = self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "div[role='alert']"))
            )
            self.logger.error(alert.text)
            return False

        try:  # Password Error
            alert = self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "div[role='alert']"))
            )
            self.logger.error(alert.text)
            return False
        except TimeoutException:
            pass
        # -- Stay Signed In
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[id='idSIButton9']"))
        ).click()
        time.sleep(1)
        self.navigate(self.url)
        if self.browser.current_url.find(".sharepoint.com") == -1:
            self.logger.info(" Xác thực thất bại!")
            return False
        self.logger.info("Authenticated")
        return True
 
    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=pd.DataFrame()
    )
    def info(self,site_url:str) -> pd.DataFrame:
        """ Lấy danh sách file và folder từ 'site_url' trên SharePoint.
        Args:
            site_url (str): Đường dẫn URL của site SharePoint cần lấy thông tin.
        Returns:
            pd.DataFrame: DataFrame chứa thông tin các file và folder khớp với mẫu, 
        """
        self.logger.info(f"Get info {site_url}")
        result = []
        self.navigate(site_url)
        rows = self.wait.until(
            method = EC.any_of(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[id^='virtualized-list'][role='row']")),
            )
        )
        for row in rows:
            # -- Type
            type_el = row.find_element(
                by = By.CSS_SELECTOR,
                value="div[data-automationid='field-DocIcon']"
            )
            try:
                type_name = type_el.find_element(By.TAG_NAME,'img').get_attribute('alt')
            except NoSuchElementException:
                type_name = None
            # -- Name
            name_el = row.find_element(
                by = By.CSS_SELECTOR,
                value="div[data-automationid='field-LinkFilename']"
            )
            name = name_el.text
            name = re.sub(self.__class__.redundant_name,'',name)
            # Agg
            result.append({
                "Type":type_name,
                "Name":name,
            })
        return pd.DataFrame(result)

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=[(None,None,"Download Error"),]
    )
    def download(self,site_url:str,file_pattern:str) -> List[Tuple[str | None,str,str]]:
        self.logger.info(f"Search {file_pattern} - {site_url}")
        result = []
        self.navigate(site_url)
        # Folder
        folder_found = False
        folders: List[str] = file_pattern.split("/")[:-1]
        for step in folders:
            time.sleep(self.retry_interval)
            gridcells = self.wait.until(
                EC.any_of(
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automationid='field-LinkFilename']")),
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automation-key^='displayNameColumn']")),
                )
            )
            for gridcell in gridcells:
                text = re.sub(self.__class__.redundant_name,'',gridcell.text)
                if re.match(step,text):
                    button = gridcell.find_element(
                        By.XPATH,
                        './/button | .//span[@role="button"]'
                    )
                    self.wait.until(EC.element_to_be_clickable(button)).click()
                    time.sleep(self.timeout)
                    folder_found = True
                    break
            time.sleep(self.retry_interval)  
        if not folder_found:
            raise LookupError(f"Folder Not Found: {file_pattern}")
        # File
        file: str = file_pattern.split("/")[-1]
        gridcells = self.wait.until(
            EC.any_of(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automationid='field-LinkFilename']")),
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automation-key^='displayNameColumn']")),
            )
        )
        for gridcell in gridcells:
            file_name = re.sub(self.__class__.redundant_name, '', gridcell.text)
            if re.match(file,file_name):
                button = gridcell.find_element(
                    By.XPATH,
                    './/button | .//span[@role="button"]'
                )
                self.wait.until(EC.element_to_be_clickable(button))
                time.sleep(self.retry_interval)
                # Copy Link
                link = None
                button.click()
                time.sleep(2)
                new_window = self.browser.window_handles[-1]  
                if new_window != self.root_window:
                    self.browser.switch_to.window(new_window)
                    while self.browser.execute_script("return document.readyState") != "complete":
                        time.sleep(self.retry_interval)
                    link = unquote(self.browser.current_url)
                    self.browser.close()
                    self.browser.switch_to.window(self.root_window)
                else:
                    close_button = self.wait.until(
                        EC.element_to_be_clickable((By.CSS_SELECTOR,"button[id='closeCommand']"))
                    )
                    time.sleep(self.retry_interval)
                    close_button.click()
                    time.sleep(self.timeout)
                # Download File
                ActionChains(self.browser).context_click(button).perform()
                download_btn = self.wait.until(
                    EC.presence_of_element_located(
                        (By.CSS_SELECTOR,"button[data-automationid='downloadCommand']")
                    )
                )
                self.wait.until(EC.element_to_be_clickable(download_btn)).click()
                time.sleep(5)
                file_path, status = self.wait_for_download_to_finish()
                self.logger.info(f"Download {file_name}: {file_path if not status else status},")
                result.append((link,os.path.join(self.download_directory,file_path),status))
        return result
    
__all__ = ["SharePoint"]



import os
import re
import time
from pathlib import Path
from typing import List, Literal

from pywinauto import Application
from pywinauto.application import WindowSpecification
from pywinauto.timings import wait_until_passes
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from NSK.bot.base import IBot
from NSK.common import retry_if_exception


class MailDealer(IBot):
    def __init__(self, url: str, username: str, password: str, **kwargs):
        super().__init__(**kwargs)
        self.url = url
        self.authenticated = self._authentication(username, password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")
    @retry_if_exception(
        failure_return=False
    )
    def _authentication(self,username:str,password:str) -> bool:
        self.navigate(self.url)
        self.wait.until(
            method = EC.element_to_be_clickable((By.ID, "fUName")),
            message = "Not Found Username Field",
        ).send_keys(username)
        self.wait.until(
            method = EC.element_to_be_clickable((By.ID, "fPassword")),
            message="Not Found Password Field",
        ).send_keys(password)
        self.wait.until(
            method=EC.element_to_be_clickable((By.CSS_SELECTOR, "input[value='ログイン']")),
            message="Not Found Login Button",
        ).click()
        time.sleep(self.retry_interval)
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        try:
            error = self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, "div[class='d_error_area ']"),
                ),
            )
            self.logger.error(error.text)
            return False
        except TimeoutException:
            if self.browser.current_url.find("app") != -1:
                self.logger.info("Authenticated")
                return True
            return True

    @retry_if_exception(
        exceptions=(
            ElementClickInterceptedException,
            StaleElementReferenceException,
            TimeoutException,
        ),
        failure_return=False
    )
    def send_mail(self,fr:str,to:str,subject:str,content:str,attachments:List[str] = [],mode:Literal["送信確認","承認依頼確認","一時保存"] = "一時保存") -> bool:
        for window in self.browser.window_handles:
            if window != self.root_window:
                self.browser.switch_to.window(window)
                self.browser.close()
        self.browser.switch_to.window(self.root_window)
        self.logger.info(f"Send mail from {fr} to {to} (subject: {subject})")
        self.wait.until(
            method = EC.frame_to_be_available_and_switch_to_it((By.CSS_SELECTOR, "iframe[id='ifmSide']")),
            message = "Not Found Side Frame",
        )
        self.wait.until(
            method = EC.element_to_be_clickable((By.CSS_SELECTOR, "span[title='メール作成']")),
            message="Not Found Create Mail Button",
        ).click()
        while len(self.browser.window_handles) == 1:
            continue
        new_window = self.browser.window_handles[-1]
        self.browser.switch_to.window(new_window)
        self.browser.maximize_window()
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        self.wait.until(
            EC.element_to_be_clickable((By.XPATH, "//button[text()='次へ']"))
        ).click()
        # From
        fr_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='fFrom']"))
        )
        fr_input.clear()
        fr_input.send_keys(fr)
        # To
        to_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='fTo[]']"))
        )
        to_input.clear()
        to_input.send_keys(to)
        # Subject
        subject_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR,"input[id='fSubject']"))
        )
        subject_input.clear()
        subject_input.send_keys(subject)
        # Body
        self.wait.until(
            EC.frame_to_be_available_and_switch_to_it((By.TAG_NAME, "iframe"))
        )
        body = self.wait.until(EC.presence_of_element_located((By.TAG_NAME, "body")))
        self.browser.execute_script(
            "arguments[0].innerText = arguments[1];", body, content
        )
        self.browser.switch_to.default_content()
        for attachment in attachments:
            # Check
            if not os.path.exists(attachment):
                raise FileNotFoundError(f"Attachment Not Found: {attachment}")
            # Attachment
            self.wait.until(
                method = EC.element_to_be_clickable((By.CSS_SELECTOR, "button[title='添付ファイル']")),
                message="Not Found Attachment Button",
            ).click()
            file_uploader: WindowSpecification = wait_until_passes(
                timeout=self.timeout,
                retry_interval=self.retry_interval,
                func=lambda: Application()
                .connect(title="Open")
                .window(title="Open"),
            )
            wait_until_passes(
                timeout=self.timeout,
                retry_interval=self.retry_interval,
                func=lambda dialog=file_uploader: dialog 
                .child_window(class_name="ComboBoxEx32")
                .child_window(class_name="Edit")
                .wrapper_object(),
            ).type_keys(re.sub(r"([{}^%~()])", r"{\1}", str(Path(attachment))),with_spaces=True)
            wait_until_passes(
                timeout=self.timeout,
                retry_interval=1,
                func=lambda dialog=file_uploader: dialog.child_window(
                    title="&Open", class_name="Button"
                ).wrapper_object(),
            ).click_input()
            for button in self.wait.until(
                method = EC.presence_of_all_elements_located((By.CSS_SELECTOR,"button[class='accent-btn__btn']")),
                message="Not Found Send Button"
            ):
                if button.text.find(mode) != -1:
                    self.browser.execute_script(
                        "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                        button,
                    )
                    time.sleep(1)
                    self.wait.until(EC.element_to_be_clickable(button)).click()
                    time.sleep(1)
                    snackbar = self.wait.until(
                        method = EC.presence_of_element_located((By.CSS_SELECTOR, "div[class='snackbar__msg']")),
                        message= "Snackbar Not Found ", 
                    )
                    self.logger.info(snackbar.text)
                    break
        self.browser.close()
        self.browser.switch_to.window(self.root_window)
        self.logger.info("Send mail successfully")
        return True



import logging
import os
import sys
import time
from abc import ABC, abstractmethod
from typing import Literal

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.ui import WebDriverWait

from .meta import IBotMeta


class IBot(ABC, metaclass=IBotMeta):
    def __init__(
        self,
        options: Options =webdriver.ChromeOptions(),
        service: Service =webdriver.ChromeService(),
        keep_alive: bool = False,
        timeout: float = 5,
        retry_interval: float = 0.5,
        log_name: str = __name__,
        log_format:str = "%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        log_level: Literal["DEBUG","INFO","WARNING","ERROR","CRITICAL"] = "INFO",
        log_file: str | None = None
    ):
        """
        Khởi tạo bot tự động hóa sử dụng trình duyệt Chrome với cấu hình logging tùy chỉnh.
        Args:
            options (Options, optional): Các tùy chọn cấu hình cho trình duyệt Chrome. Mặc định là ChromeOptions() mặc định.
            service (Service, optional): Đối tượng service điều khiển ChromeDriver. Mặc định là ChromeService() mặc định.
            keep_alive (bool, optional): Giữ kết nối trình duyệt sống sau khi script kết thúc hay không. Mặc định là False.
            timeout (float, optional): Thời gian timeout (giây) chờ các thao tác Selenium. Mặc định là 5 giây.
            retry_interval (float, optional): Khoảng thời gian chờ giữa các lần thử lại thao tác (polling interval). Mặc định là 0.5 giây.
            log_name (str, optional): Tên logger sử dụng để phân biệt nguồn log (thường là tên class/module). Mặc định là tên module hiện tại (__name__).
            log_format (str, optional): Định dạng của dòng log, theo chuẩn của logging.Formatter. Mặc định là thời gian - cấp độ - tên logger - message.
            log_level (Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], optional): Mức độ log mong muốn. Mặc định là "INFO".
            log_file (str | None, optional): Đường dẫn file để ghi log. Nếu là None thì chỉ in log ra console. Mặc định là None.
        Note:
            Cấu hình logging của các IBot sẽ bị ghi đè nếu ứng dụng chính hoặc module cha đã thiết lập cấu hình logging riêng trước đó.
        """
        # ------ # 
        self.timeout = timeout
        self.authenticated = False
        self.retry_interval = retry_interval
        # ------ # 
        self.options = options
        self.service = service
        self.browser = webdriver.Chrome(
            options=self.options,
            service=self.service,
            keep_alive=keep_alive,
        )
        self.browser.maximize_window()
        self.wait = WebDriverWait(
            driver=self.browser,
            timeout=self.timeout,
            poll_frequency=self.retry_interval,
        )
        self.root_window = self.browser.window_handles[0]
        # ------ # 
        self.logger = logging.getLogger(log_name)
        if not self.logger.hasHandlers():
            self.logger.setLevel(log_level)
            formatter = logging.Formatter(log_format)
            # -- Console Logging
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)
            # -- File Logging
            if log_file:
                os.makedirs(os.path.dirname(os.path.abspath(log_file)), exist_ok=True)
                file_handler = logging.FileHandler(log_file, encoding="utf-8")
                file_handler.setFormatter(formatter)
                self.logger.addHandler(file_handler)
        # ------ # 
        self.download_directory = self.options.experimental_options.get(
            "prefs", {}
        ).get(
            "download.default_directory",
            os.path.join(os.path.expanduser("~"), "downloads"),
        )
        os.makedirs(self.download_directory, exist_ok=True)
        # ------ # 
        self.logger.info(f"Initiating {self.__class__.__name__}")

    def __del__(self):
        try:
            self.browser.quit()
        except Exception:
            pass
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.browser.quit()
        
    def navigate(self, url, wait_for_complete: bool = True):
        time.sleep(self.retry_interval)
        self.browser.execute_script("window.stop()")
        time.sleep(self.retry_interval)
        self.browser.get(url)
        time.sleep(self.retry_interval)
        if wait_for_complete:
            while (
                self.browser.execute_script("return document.readyState") != "complete"
            ):
                time.sleep(self.retry_interval)
        time.sleep(self.retry_interval)

    @abstractmethod
    def _authentication(self, **kwargs) -> bool:
        """Abstract method to implement authentication"""
        pass

    def open_new_tab(self) -> int:
        before_windows = self.browser.window_handles
        self.browser.execute_script("window.open('');")
        after_windows = self.browser.window_handles
        return list(set(after_windows) - set(before_windows))[0]

    def wait_for_download_to_start(self) -> list[str]:
        while True:
            downloading_files = [
                filename
                for filename in os.listdir(self.download_directory)
                if filename.endswith(".crdownload")
            ]
            if downloading_files:
                return downloading_files

    def wait_for_download_to_finish(self) -> tuple[str, str]:
        window_id = self.open_new_tab()
        self.browser.switch_to.window(window_id)
        self.navigate("chrome://downloads")
        download_items: list[WebElement] = self.browser.execute_script(
            """
            return document.
                querySelector("downloads-manager").shadowRoot
                .querySelector("#mainContainer #downloadsList #list")
                .querySelectorAll("downloads-item")
        """
        )
        item_id = download_items[0].get_attribute("id")
        while self.browser.execute_script(
            f"""
            return document
                .querySelector("downloads-manager").shadowRoot
                .querySelector("#downloadsList #list")
                .querySelector("#{item_id}").shadowRoot
                .querySelector("#content #details #progress")
            """
        ):  # Progess
            time.sleep(self.retry_interval)
        name = self.browser.execute_script(
            f"""
            return document
                .querySelector("downloads-manager").shadowRoot
                .querySelector("#downloadsList")
                .querySelector("#list")
                .querySelector("#{item_id}").shadowRoot
                .querySelector("#content")
                .querySelector("#details")
                .querySelector("#title-area")
                .querySelector("#name")
                .getAttribute("title")
            """
        )
        tag = self.browser.execute_script(
            f"""
            return document
                .querySelector("downloads-manager").shadowRoot
                .querySelector("#downloadsList")
                .querySelector("#list")
                .querySelector("#{item_id}").shadowRoot
                .querySelector("#content")
                .querySelector("#details")
                .querySelector("#title-area")
                .querySelector("#tag")
                .textContent.trim();
            """
        )
        self.browser.close()
        self.browser.switch_to.window(self.root_window)
        return name, tag